/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/*!
@file movingEdgesDetector.cpp
@ingroup ExamplePrograms
@brief A simple moving edge detector based on the Canny edge detector.

This is an example application which combines several workers from the QVision, and a user-defined one, to perform detection of moving edges on an input video sequence.

@note The access operator () used in the instructions @code if (movementImage(edgePoint) > 10) @endcode and @code movingBordersImage(edgePoint) = 255; @endcode to obtain and set image pixel values is not the most efficient way of accessing image data provided in the QVision. The final application offers a slow frame rate for that reason. See section @ref FastImageAccess for more info about speeding up performance on image access, which could make this example application run much faster.


@section Usage
Once compiled, the program should be used with this command line:
\code
	./movingEdgesDetector --URL=<path or url to video>
\endcode

@param URL Video source identificator. See dynamic property URL for @ref QVMPlayerCamera.
@author PARP Research Group. University of Murcia, Spain.
*/

// Comment this if you don't want inpainting.
#ifndef DOXYGEN_IGNORE_THIS

#include <QVApplication>
#include <QVMPlayerCamera>
#include <QVDefaultGUI>
#include <QVImageCanvas>
#include <QVCannyEdgeDetector>
#include <QVImageRetarderWorker>
#include <qvippworkers.h>

///////////////////////////////////////////////////
#include <QVPolyline>
class MovingEdgesDetector: public QVWorker
	{
	public:
		MovingEdgesDetector(QString name): QVWorker(name)
			{
			addProperty< QList<QVPolyline> >("Image borders", inputFlag);
			addProperty< QVImage<uChar,1> >("Movement image", inputFlag);
			addProperty< QVImage<uChar,1> >("Moving borders image", outputFlag);
			}

		void iterate()
			{
			// Read parameters
			const QList< QVPolyline > imageBorders = getPropertyValue< QList<QVPolyline> >("Image borders");
			const QVImage<uChar,1> movementImage = getPropertyValue< QVImage<uChar,1> >("Movement image");

			// Create and initialize to zero the output image, with the same dimensions as 'movementImage'
			QVImage<uChar,1 > movingBordersImage(movementImage.getCols(), movementImage.getRows());
			Set(0, movingBordersImage);

			// Transverse edge point lists, checking if their location at the movement image has a greyscale value
			// greater than 10. If so, set that location at image 'movingBordersImage' to 255.
			foreach(QVPolyline edge, imageBorders)
				foreach(QPoint edgePoint, edge)
					if (movementImage(edgePoint) > 10)
						movingBordersImage(edgePoint) = 255;

			// Store resulting image
			setPropertyValue< QVImage<uChar,1> >("Moving borders image", movingBordersImage);
			timeFlag("Publish resulting images");
			}
	};

int main(int argc, char *argv[])
        {
        QVApplication app(argc, argv,
                "Example program for QVision library. Obtains several features from input video frames."
                );

        QVDefaultGUI interface;

        QVMPlayerCamera camera("Video");

        QVCannyEdgeDetector cannyWorker("Canny worker");
        camera.link(&cannyWorker, "Input image");

	QVImageCanvas cannyDisplayer("Canny");
	cannyDisplayer.linkProperty(cannyWorker,"Input image");
	cannyDisplayer.linkProperty(cannyWorker,"Output contours");

	QVImageRetarderWorker<uChar,1> retarderWorker("Image retarder worker");
	camera.link(&retarderWorker, "Input image");

	QVIPPAbsDiffWorker<uChar,1> absDiffWorker("Absolute difference worker");
	camera.link(&absDiffWorker, "src1");
	retarderWorker.linkProperty("Output image", &absDiffWorker, "src2", QVWorker::SynchronousLink);

	QVImageCanvas movementDisplayer("Movement detector");
	movementDisplayer.linkProperty(absDiffWorker,"dest");
	
	///////////////////////
	MovingEdgesDetector movingEdgesDetector("Moving edges detector");
	cannyWorker.linkProperty("Output contours", &movingEdgesDetector, "Image borders", QVWorker::SynchronousLink);
	absDiffWorker.linkProperty("dest", &movingEdgesDetector, "Movement image", QVWorker::SynchronousLink);

	QVImageCanvas edgeMovementDisplayer("Movement detector");
	edgeMovementDisplayer.linkProperty(movingEdgesDetector,"Moving borders image");

        return app.exec();
        }

#endif
