/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvglcanvas.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <QMouseEvent>

#include <qvipp.h>
#include <qvdefines.h>
#include <qvmatrixalgebra.h>

#include <qvgui/qvglcanvas.h>

#define GL_VIEW_ASPECT   1.333

QVGLCanvas::QVGLCanvas(const QString &title, QWidget* parent):
	QGLWidget(parent), maxwide(2.0), cx(0.0), cy(0.0), cz(0.0),
	trackballQuat(QVQuaternion::trackball(0.0, 0.0, 0.0, 0.0)),
	zoom(45), pressedleft(FALSE), pressedright(FALSE)
	{
	resize(400,(int)(400/GL_VIEW_ASPECT));
	setWindowTitle(title);
	show();
	}

QVGLCanvas::~QVGLCanvas ()
	{
	// FIXME: is it needed?: makeCurrent(); 
	}

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////   OpenGL related:     ////////////////////////////
////////////////////////////////////////////////////////////////////////////////
void QVGLCanvas::linkModelMatrix(QVWorker &worker, const QString &matrixPropertyName, const QString &objectName)
	{
	/// @todo check that worker property is a QVMatrix
	worker.linkProperty(matrixPropertyName, this, objectName + " matrix", QVWorker::AsynchronousLink);

	/// @todo decidir si esto debería ir aqui. Esto actualiza el canvas gráfico cada vez que las propiedades del worker se actualizan.
	QObject::connect(&worker, SIGNAL(endIteration(uint, int)), this, SLOT(updateGL()));
	}

void QVGLCanvas::add(const QV3DModel &model, const QString &name)
	{
	addProperty<QV3DModel>(name, inputFlag | outputFlag, model, name + " 3d object model");
	addProperty<QVMatrix>(name + " matrix", inputFlag | outputFlag, QVMatrix::identity(4), name + " 3d location matrix");
	}

void QVGLCanvas::initializeGL()
	{
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	glDisable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	
	glEnable(GL_DITHER);
	glShadeModel(GL_SMOOTH);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
	glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);
	
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST); 
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST); 
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_TEXTURE_2D);

	init();
	}

void QVGLCanvas::drawModels()
	{
	readInputProperties();

	glScaled((GLdouble)1/maxwide,(GLdouble)1/maxwide,(GLdouble)1/maxwide);

	foreach(QString name, getPropertyListByType<QV3DModel>())
		{
		QV3DModel model = getPropertyValue<QV3DModel>(name);
		QVMatrix matrix = getPropertyValue<QVMatrix>(name + " matrix");

		//std::cout << "maxwide = " << maxwide << std::endl;
		//std::cout << "Matrix = " << matrix << std::endl;
		//std::cout << "Model: '" << qPrintable(name) << "'." << std::endl;
		if (matrix.getCols() == 4 && matrix.getRows() == 4)
			{
			glPushMatrix();

			/// @todo Aplicar matriz de translación/rotación conjunta.
			GLfloat m[4][4];
			QVMatrix M = pseudoInverse(matrix);

			for (int i=0; i <4; i++)
				for (int j=0; j <4; j++)
					m[i][j] = matrix(i,j);

			for (int i=0; i <3; i++)
				{ m[3][i] = 0; m[i][3] = 0; }

			glTranslated(-M(0,3), -M(1,3), -M(2,3));

			glMultMatrixf(&m[0][0]);
			
			model.paint(*this);
			glPopMatrix();

			glFlush();
			}
		else
			model.paint(*this);
		}
	}

void QVGLCanvas::paintGL()
	{
	glMatrixMode(GL_PROJECTION);

	glLoadIdentity();
	gluPerspective(zoom,(float)size().width()/(float)size().height(),1,100);
	glMatrixMode(GL_MODELVIEW);

	glLoadIdentity();
	
	glTranslatef(0,0,-4); 
	glScaled((GLdouble)1/maxwide,(GLdouble)1/maxwide,(GLdouble)1/maxwide);


	GLfloat m[4][4];
	QVMatrix M(trackballQuat);

	for (int i=0; i <4; i++)
		for (int j=0; j <4; j++)
			m[i][j] = M(i,j);

	glMultMatrixf(&m[0][0]);
	
	glRotatef(270,1,0,0); 
	
	glTranslatef(-cx,-cy,-cz); 

	glClearColor(.5, .5, .75, 1.0);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	drawModels();

	display();
	}

void QVGLCanvas::resizeGL( int w, int h )
	{
	glViewport(0,0,w,h); 
	reshape(w,h);

	drawModels();
	}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////   Mouse & keyboard:     ///////////////////////
////////////////////////////////////////////////////////////////////////////////

void QVGLCanvas::mousePressEvent(QMouseEvent *event)
{
    beginx = event->x();
    beginy = event->y();
    if(event->button() == Qt::LeftButton) {
        pressedleft = TRUE;
    } else if(event->button() == Qt::RightButton) {
        pressedright = TRUE;
    }
}

void QVGLCanvas::mouseReleaseEvent(QMouseEvent *event)
{
    if(event->button() == Qt::LeftButton) {
        pressedleft = FALSE;
    } else if(event->button() == Qt::RightButton) {
        pressedright = FALSE;
    }
}

void QVGLCanvas::mouseMoveEvent(QMouseEvent *event)
{
    int x,y;

    x = (int) event->x();
    y = (int) event->y();

    if (pressedleft) {
	QVQuaternion spinQuat = QVQuaternion::trackball(
			(2.0*beginx - size().width())	/ size().width(),
			(size().height() - 2.0*beginy)	/ size().height(),
			(2.0*x - size().width())	/ size().width(),
			(size().height() - 2.0*y)	/ size().height());

	trackballQuat = spinQuat * trackballQuat;
        updateGL();
    }

    if (pressedright) {
        /* zooming drag */
        //printf("MouseMove right button\n");
        zoom += ((y - beginy) / size().height()) * 40;
        if (zoom < 5) zoom = 5;
        if (zoom > 120) zoom = 120;
        /* zoom has changed, redraw  */
        updateGL();
    }
    beginx = x;
    beginy = y;
}

void QVGLCanvas::wheelEvent(QWheelEvent *event)
{
    /* zooming wheel */
    zoom -= event->delta()/32;
    if (zoom < 5) zoom = 5;
    if (zoom > 120) zoom = 120;
    /* zoom has changed, redraw  */
    updateGL();
}



void QVGLCanvas::keyPressEvent(QKeyEvent *event)
{
    //printf("KeyPress\n");
    switch(event->key()) {
      case Qt::Key_Left:
        cx -= maxwide/20;
        break;
      case Qt::Key_Right:
        cx += maxwide/20;
        break;
      case Qt::Key_Up:
        cy += maxwide/20;
        break;
      case Qt::Key_Down:
        cy -= maxwide/20;
        break;
      case Qt::Key_PageUp:
        cz += maxwide/20;
        break;
      case Qt::Key_PageDown:
        cz -= maxwide/20;
        break;
    }
    /* Ha cambiado el centro, redibujar: */
    updateGL();
}

void QVGLCanvas::closeEvent(QCloseEvent * event)
{
	Q_UNUSED(event);
    emit closed();
}

//////////////////////////////////////////////////////////////

int nearest_gl_texture_size(int v)
	{
	int n = 0, last = 0;
	for (int s = 0; s < 32; ++s)
		{
		if (((v>>s) & 1) == 1)
			{
			n++;
			last = s;
			}
		}

	if (n > 1)
		return 1 << (last+1);
	return 1 << last;
	}

GLuint QVGLCanvas::bindTexture(const QVImage<uChar,3> &image, GLenum target)
	{
	const uInt cols = image.getCols(), rows = image.getRows(), step = image.getStep();
	bool clean = false;

	// Scale the pixmap if needed. GL textures needs to have the
	// dimensions 2^n+2(border) x 2^m+2(border).
	int textureImage_w = nearest_gl_texture_size(cols);
	int textureImage_h = nearest_gl_texture_size(rows);

	QVImage<uChar,3> textureImage(textureImage_w, textureImage_h);
	Resize(image, textureImage);

	GLuint textureImage_id;
	glGenTextures(1, &textureImage_id);
	glBindTexture(target, textureImage_id);
	glTexParameterf(target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glHint(GL_GENERATE_MIPMAP_HINT_SGIS, GL_NICEST);
	glTexParameteri(target, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
	glTexParameterf(target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexImage2D(target, 0, GL_RGB, textureImage_w, textureImage_h, 0, GL_BGR, GL_UNSIGNED_BYTE, textureImage.getReadData());

	return textureImage_id;
	}
