/*
 *	Copyright (C) 2007, 2008. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PROJECTIVE_H
#define PROJECTIVE_H

#include <QPair>
#include <qvmath.h>
#include <qvdta.h>
#include <QVMatrix>

/// @file qvprojective.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////// HOMOGRAPHY CALIBRATION FUNCTION ///////////////////////////////////////////////

/*!
@brief Obtains an homography from two lists of corresponding points.
@ingroup qvprojectivegeometry

This function returns the homography that maps the points from a source position to a
destination position, according to a projective transormation.

@deprecated Use @ref ComputeProjectiveHomography instead
@param sourcePoints list of source points.
@param destinationPoints list of destination points.
*/
QVMatrix ComputeHomography(const QList<QPointF> &sourcePoints, const QList<QPointF> &destPoints);

/*!
@brief Obtains an homography from two lists of corresponding points.
@ingroup qvprojectivegeometry

This function returns the homography that maps the points from a source position to a
destination position, according to a projective transformation.

@param matchings list of matchings
*/
QVMatrix ComputeProjectiveHomography(const QList< QPair<QPointF, QPointF> > &matchings);

/*!
@ingroup qvprojectivegeometry
@todo write documentation for this function
*/
QVMatrix ComputeEuclideanHomography(const QPair<QPointF, QPointF> &firstMatching, const QPair<QPointF, QPointF> &secondMatching);

/*!
@ingroup qvprojectivegeometry
@todo write documentation for this function
*/
QVMatrix ComputeEuclideanHomography(const QList< QPair<QPointF, QPointF> > &matchings);

/*!
@ingroup qvprojectivegeometry
@todo write documentation for this function
*/
QPointF ApplyHomography(const QVMatrix &homography, const QPointF &point);

/*!
@ingroup qvprojectivegeometry
@todo write documentation for this function
*/
QList<QPointF> ApplyHomography(const QVMatrix &homography, const QList<QPointF> &sourcePoints);

/*!
@brief Function to test if a 3x3 matrix corresponds to an homography.
@ingroup qvprojectivegeometry

Every matrix which holds a perspective deformation from one plane to another should validate some constrains. The most important is that its two first columns
should be perpendicular, and with a similar size, because for the matrix to correspond to an homography, they should be contained in a base of a rotated coordinate system.
Given the following homography matrix:

\f$ H = \left[ R_1 R_2 -CR^t \right] \f$

The error value returned by this function for it will be:

\f$ error = \left| \frac{ \|R_1\| - \|R_2\| }{ \|R_1\| + \|R_2\| } \right| + \left| \frac{ R_1 \cdot R_2}{ \|R_1\| \|R_2\| } \right| \f$

That error is a measure of the difference of sizes between the norm of the two column vectors of the homography, corresponding to the two first columns of the rotation matrix, and their dot product. When both values are close to zero, the matrix corresponds to an homography, and it won't otherwise.

A good method to stablish that a matrix corresponds to an homography using this function, can be done testing the return value with a threshold of aproximately 0.3. If the return value of this function for a matrix is lower than this threshold, that matrix is likely to correspond to an homography, and is not likely to correspond otherwise.

@param homography a possible homography matrix.
@return a value close to 1 when the matrix doesn't corresponds to an homography, and close to 0 when it is close to be an homography.
*/
double HomographyTestError(const QVMatrix &homography);


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////// CAMERA CALIBRATION FUNCTION ///////////////////////////////////////////////////

/*
@brief Obtains intrinsic and extrinsic camera matrix decomposition from a planar <a href="http://en.wikipedia.org/wiki/Homography">homography</a>.
@ingroup qvproyective

@todo explain how focal value is needed for intrinsic matrix, and how it is searched.
This is a direct version of the function @ref GetIntrinsicCameraMatrixFromHomography, which doesn't 

@param H homography matrix
@param K matrix to store resulting intrinsic camera matrix.
@param M matrix to store resulting extrinsic camera matrix.
*/
//bool DecomposeCameraMatricesFromPlanarHomography(const QVMatrix &H, QVMatrix &K, QVMatrix &M);

/*!
@ingroup qvprojectivegeometry
@todo write documentation for this function
*/
void GetExtrinsicCameraMatrixFromHomography(const QVMatrix &K, const QVMatrix &H, QVMatrix &M4x4);

/*!
@ingroup qvprojectivegeometry
@todo write documentation for this function
*/
void GetDirectIntrinsicCameraMatrixFromHomography(const QVMatrix &H, QVMatrix &K);

/*!
@ingroup qvprojectivegeometry
@todo write documentation for this function
*/
void GetIntrinsicCameraMatrixFromHomography(const QVMatrix &H, QVMatrix &K,
	double focal = 3, const double maxFocal = 50, const int maxIterations = 100, const double maxError = 0.00001);

#endif
